#include "SceneGraph.h"
#include "Exporter.h"
#include "UnigraphicsParser.h"
#include "X3dParser.h"
#include "Preprocessor.h"
#include "ViewCell.h"
#include "Environment.h"
#include "RenderSimulator.h"

Preprocessor::Preprocessor():
mKdTree(NULL),
mBspTree(NULL),
mRenderSimulator(NULL)
{
}


Preprocessor::~Preprocessor()
{
	DeleteViewCells();
		
	DEL_PTR(mBspTree);
	DEL_PTR(mKdTree);

	DEL_PTR(mRenderSimulator);
}

bool
Preprocessor::LoadViewCells(const string filename)
{
	X3dParser parser;
	
	environment->GetFloatValue("ViewCells.height", parser.mViewCellHeight);
	
	return parser.ParseFile(filename, mViewCells);
}

bool
Preprocessor::ParseViewCellsOptions()
{
	// parse type of view cells
	char viewCellsStr[64];
	environment->GetStringValue("ViewCells.hierarchy", viewCellsStr);

	if (strcmp(viewCellsStr, "bspTree") == 0)
	{
		ViewCell::sHierarchy = ViewCell::BSP;
	}
	else if (strcmp(viewCellsStr, "kdTree") == 0)
	{
		ViewCell::sHierarchy = ViewCell::KD;
	}
	else if (strcmp(viewCellsStr, "vspTree") == 0)
	{
		ViewCell::sHierarchy = ViewCell::VSP;
	}
	else if (strcmp(viewCellsStr, "sceneDependent") == 0)
	{
		//TODO
	}
	else
	{
		cerr<<"Wrong view cells type" << viewCellsStr << endl;
		exit(1);
	}

	return true;
}

RenderSimulator *Preprocessor::GetRenderSimulator()
{
	if (mRenderSimulator)
		return mRenderSimulator;

    float objRenderCost = 0, vcOverhead = 0, moveSpeed = 0;

	environment->GetFloatValue("Simulation.objRenderCost",objRenderCost);
	environment->GetFloatValue("Simulation.vcOverhead", vcOverhead);
	environment->GetFloatValue("Simulation.moveSpeed", moveSpeed);

	Debug << "render simulator using render cost=" << objRenderCost << ", vc overhead=" << vcOverhead << ", move speed=" << moveSpeed << endl;
	if (ViewCell::sHierarchy == ViewCell::BSP)
	{
		mRenderSimulator = new BspViewCellRenderSimulator(objRenderCost, vcOverhead, moveSpeed, mBspTree);
		Debug << "creating bsp render simulator" << endl;
	}
	else if (ViewCell::sHierarchy == ViewCell::KD)// KD view cells
	{
		mRenderSimulator = new KdViewCellRenderSimulator(objRenderCost, vcOverhead, moveSpeed, mKdTree);
		Debug << "creating kd render simulator" << endl;
	}
	else
	{
		Debug << "not implemented yet" << endl;
		return NULL;
	}

	return mRenderSimulator;
}

void Preprocessor::DeleteViewCells()
{
	for (int i = 0; i < (int)mViewCells.size(); ++ i)
	{
		Mesh *mesh = mViewCells[i]->GetMesh();
		DEL_PTR(mesh);
	}
	CLEAR_CONTAINER(mViewCells);
}

int
SplitFilenames(const string str, vector<string> &filenames)
{
	int pos = 0;

	while(1) {
		int npos = str.find(';', pos);
		
		if (npos < 0 || npos - pos < 1)
			break;
		filenames.push_back(string(str, pos, npos - pos));
		pos = npos + 1;
	}
	
	filenames.push_back(string(str, pos, str.size() - pos));
	return filenames.size();
}

bool
Preprocessor::LoadScene(const string filename)
{
  // use leaf nodes of the original spatial hiearrchy as occludees

  mSceneGraph = new SceneGraph;

  
  Parser *parser;
	vector<string> filenames;
	int files = SplitFilenames(filename, filenames);
	cout<<files<<endl;
	bool result = false;
	if (files == 1) {
		
		if (strstr(filename.c_str(), ".x3d"))
			parser = new X3dParser;
		else
			parser = new UnigraphicsParser;

		cout<<filename<<endl;
		result = parser->ParseFile(filename, &mSceneGraph->mRoot);

		delete parser;

	} else {
		// root for different files
		mSceneGraph->mRoot = new SceneGraphNode;
		for (int i= 0; i < filenames.size(); i++) {
			if (strstr(filenames[i].c_str(), ".x3d"))
				parser = new X3dParser;
			else
				parser = new UnigraphicsParser;
			
			SceneGraphNode *node;
			if (parser->ParseFile(filenames[i], &node)) {
				mSceneGraph->mRoot->mChildren.push_back(node);
				// at least one file parsed
				result = true;
			}
			delete parser;
		}
	}
	

	if (result) {
		mSceneGraph->AssignObjectIds();
		int intersectables, faces;
		mSceneGraph->GetStatistics(intersectables, faces);
		cout<<filename<<" parsed successfully."<<endl;
		cout<<"#NUM_OBJECTS (Total numner of objects)\n"<<intersectables<<endl;
		cout<<"#NUM_FACES (Total numner of faces)\n"<<faces<<endl;
	}

	
  return result;
}

bool
Preprocessor::ExportPreprocessedData(const string filename)
{
  return false;
}

bool
Preprocessor::BuildKdTree()
{
  mKdTree = new KdTree;
  // add mesh instances of the scene graph to the root of the tree
  KdLeaf *root = (KdLeaf *)mKdTree->GetRoot();
  mSceneGraph->CollectObjects(&root->mObjects);
  
  mKdTree->Construct();
  return true;
}

void
Preprocessor::KdTreeStatistics(ostream &s)
{
  s<<mKdTree->GetStatistics();
}

void
Preprocessor::BspTreeStatistics(ostream &s)
{
	s << mBspTree->GetStatistics();
}

bool
Preprocessor::Export( const string filename,
											const bool scene,
											const bool kdtree,
											const bool bsptree
											)
{
  Exporter *exporter = Exporter::GetExporter(filename);
	
  if (exporter) {
    if (scene)
      exporter->ExportScene(mSceneGraph->mRoot);

    if (kdtree) {
      exporter->SetWireframe();
      exporter->ExportKdTree(*mKdTree);
    }

	if (bsptree) {
		//exporter->SetWireframe();
		exporter->ExportBspTree(*mBspTree);
	}

    delete exporter;
    return true;
  }

  return false;
}
